from __future__ import annotations
from richard.entity.BlockResult import BlockResult
from richard.entity.ProcessResult import ProcessResult
from richard.entity.SentenceRequest import SentenceRequest
from richard.interface.SomeProcessor import SomeProcessor


class ControlBlock:

    processor: SomeProcessor
    next_block: ControlBlock


    def __init__(self, processor) -> None:
        self.processor = processor
    

    def process(self, request: SentenceRequest) -> BlockResult:
        pass

    


from richard.entity.BlockResult import BlockResult
from richard.entity.SentenceRequest import SentenceRequest
from richard.entity.ControlBlock import ControlBlock


class FindOne(ControlBlock):
    """
    This block goes through all alternative products until it finds one that gives no error.
    If no such product could be found, the first error is returned.
    """
    

    def process(self, request: SentenceRequest) -> BlockResult:
        result = self.processor.process(request)
        if result.error_code != '':
            return BlockResult(result.error_code, result.error_args)

        request.set_alternative_products(self.processor, result.products)

        error_code = ""
        error_args = []

        for product in result.products:

            request.set_current_product(self.processor, product)
          
            next_block_result = self.next_block.process(request)
            if next_block_result.error_code == "":
               # first successful product found: quit
               return BlockResult('', [])
            else:
                error_code = next_block_result.error_code
                error_args = next_block_result.error_args
            
        return BlockResult(error_code, error_args)

    


import unittest

from richard.constants import POS_TYPE_RELATION, POS_TYPE_WORD_FORM
from richard.entity.GrammarRule import GrammarRule
from richard.entity.GrammarRules import GrammarRules
from richard.entity.Log import Log
from richard.entity.RuleConstituent import RuleConstituent
from richard.processor.parser.earley.EarleyParser import EarleyParser

class TestEarleyParser(unittest.TestCase):
   
    def test_ambiguity(self):
        # create a toy grammar to generates superfluous sentences (to test the parsing of ambiguity)
        grammarRules = GrammarRules([
            # s(P1) -> np(E1) vp(P1, E2)
            GrammarRule(
                RuleConstituent("s", ["P1"], POS_TYPE_RELATION),
                [RuleConstituent("np", ["E1"], POS_TYPE_RELATION), RuleConstituent("vp", ["P1", "E2"], POS_TYPE_RELATION)],
                lambda sem: sem
            ),
            # vp(P1, E1) -> verb(P1) np(E1)
            GrammarRule(
                RuleConstituent("vp", ["P1", "E1"], POS_TYPE_RELATION),
                [RuleConstituent("verb", ["P1"], POS_TYPE_RELATION), RuleConstituent("np", ["E1"], POS_TYPE_RELATION)],
                lambda sem: sem
            ),
            # superfluous: s(P1) -> np(E1) verb(P1) np(E2)
            GrammarRule(
                RuleConstituent("s", ["P1"], POS_TYPE_RELATION),
                [RuleConstituent("np", ["E1"], POS_TYPE_RELATION), RuleConstituent("verb", ["P1"], POS_TYPE_RELATION), RuleConstituent("np", ["E2"], POS_TYPE_RELATION)],
                lambda sem: sem
            ),
            # np(E1) -> noun(E1)
            GrammarRule(
                RuleConstituent("np", ["E1"], POS_TYPE_RELATION),
                [RuleConstituent("noun", ["E1"], POS_TYPE_RELATION)],
                lambda sem: sem
            ),
            # noun(E1) -> proper_noun(E1)
            GrammarRule(
                RuleConstituent("noun", ["E1"], POS_TYPE_RELATION),
                [RuleConstituent("proper_noun", ["E1"], POS_TYPE_RELATION)],
                lambda sem: sem
            ),
            # verb(P1) -> "loves"
            GrammarRule(
                RuleConstituent("verb", ["P1"], POS_TYPE_RELATION),
                [RuleConstituent("loves", [], POS_TYPE_WORD_FORM)],
                lambda sem: sem
            ),
            # proper_noun(P1) -> "John"
            GrammarRule(
                RuleConstituent("proper_noun", ["E1"], POS_TYPE_RELATION),
                [RuleConstituent("john", [], POS_TYPE_WORD_FORM)],
                lambda sem: sem
            ),
            # proper_noun(P1) -> "Mary"
            GrammarRule(
                RuleConstituent("proper_noun", ["E1"], POS_TYPE_RELATION),
                [RuleConstituent("mary", [], POS_TYPE_WORD_FORM)],
                lambda sem: sem
            ),            
            # superfluous: noun(P1) -> "John"
            GrammarRule(
                RuleConstituent("noun", ["E1"], POS_TYPE_RELATION),
                [RuleConstituent("john", [], POS_TYPE_WORD_FORM)],
                lambda sem: sem
            ),
        ])
        parser = EarleyParser()
        result = parser.parse(grammarRules, ["John", "loves", "Mary"])
                    
        self.assertEqual(len(result.products), 4)
        self.assertEqual(result.products[0].inline_str(), "s(np(noun(john 'John')) verb(loves 'loves') np(noun(proper_noun(mary 'Mary'))))")
        self.assertEqual(result.products[1].inline_str(), "s(np(noun(proper_noun(john 'John'))) verb(loves 'loves') np(noun(proper_noun(mary 'Mary'))))")
        self.assertEqual(result.products[2].inline_str(), "s(np(noun(john 'John')) vp(verb(loves 'loves') np(noun(proper_noun(mary 'Mary')))))")
        self.assertEqual(result.products[3].inline_str(), "s(np(noun(proper_noun(john 'John'))) vp(verb(loves 'loves') np(noun(proper_noun(mary 'Mary')))))")




class Log:
    active: bool = True

    def __init__(self, active: bool):
        self.active = False

    def is_active(self):
        return self.active
    
    def add_debug(self, code: str, text: str):
        print(code + ": " + text)




import unittest

from richard.Domain import Domain
from richard.Pipeline import Pipeline
from richard.block.FindOne import FindOne
from richard.entity.Entity import Entity
from richard.entity.Relation import Relation
from richard.store.Record import Record
from richard.entity.SentenceRequest import SentenceRequest
from richard.processor.parser.BasicParser import BasicParser
from richard.processor.semantic_composer.SemanticComposer import SemanticComposer
from richard.processor.semantic_executor.SemanticExecutor import SemanticExecutor
from richard.processor.tokenizer.BasicTokenizer import BasicTokenizer
from richard.semantics.commands import find
from richard.store.MemoryDb import MemoryDb


class TestChat80(unittest.TestCase):
    """
    Mimics a Chat80 dialog (found here: https://github.com/JanWielemaker/chat80/blob/master/prolog/chat80/demo)
    Topics:
    - aggregation
    - extraposition (long distance despendencies)
    - relative clauses
    """
   
    def test_chat80(self):

        # database

        db = MemoryDb()
        db.insert(Record('river', {'id': 'amazon'}))
        db.insert(Record('river', {'id': 'brahmaputra'}))
        
        db.insert(Record('country', {'id': 'afghanistan'}))
        db.insert(Record('country', {'id': 'china'}))

        db.insert(Record('borders', {'county_id1': 'afghanistan', 'country_id2': 'china'}))

        # domain

        domain = Domain([
            Entity("river", lambda: db.select(Record('river')).field('id')),
            Entity("country", lambda: db.select(Record('country')).field('id')),
        ], [
            Relation("in_continent", ['place', 'continent'], lambda parent, child: db.select(Record('has_child', {'parent': parent, 'child': child}))),
            Relation("borders", ['country', 'country'], lambda country1, country2: db.select(Record('borders', {'county_id1': country1, 'county_id1': country2}))),
        ])

        # grammar

        grammar = [
            { 
                "syn": "s -> 'what' nbar 'are' 'there' '?'", 
                "sem": lambda nbar: lambda: nbar()
            },
            { 
                "syn": "s -> 'does' np vp_no_sub '?'", 
                "sem": lambda np, vp_no_sub: lambda: find(np(), vp_no_sub)
            },
            { 
                "syn": "vp_no_sub -> verb np", 
                "sem": lambda verb, np:
                        lambda subject: find(
                            np(), verb)
            },
            { "syn": "verb -> 'border'", "sem": lambda: lambda subject, object: domain.relation_exists('border', [subject, object]) },
            { "syn": "nbar -> noun", "sem": lambda noun: lambda: noun() },
            { "syn": "noun -> proper_noun", "sem": lambda proper_noun: lambda: None }, # todo
            { "syn": "noun -> 'rivers'", "sem": lambda: lambda: domain.get_entity_ids('river') },
        ]

        # pipeline

        tokenizer = BasicTokenizer()
        parser = BasicParser(grammar, tokenizer)
        composer = SemanticComposer(parser)
        executor = SemanticExecutor(composer)

        pipeline = Pipeline([
            FindOne(tokenizer),
            FindOne(parser),
            FindOne(composer),
            FindOne(executor)
        ])

        # testing

        tests = [
            ["What rivers are there?", ['amazon', 'brahmaputra']],
            ["Does Afghanistan border China?", 'yes']
        ]

        for test in tests:
            question, answer = test
            request = SentenceRequest(question)
            result = pipeline.enter(request)

            if not result.error_code == "":
                print(result.error_code, result.error_args) 
                break

            results = executor.get_results(request)
            self.assertEqual(set(answer), set(results))



from __future__ import annotations
from dataclasses import dataclass, field


@dataclass(frozen=True)
class Record:
    """
    A record represents a named tuple (row) in relation (table) of a database. This may be any type of database.
    """

    table: str
    # default: empty dictionary
    values: dict[str, int|float|str] = field(default_factory=dict)


    def __post_init__(self):
        for name, value in self.values.items():
            if not name.isidentifier():
                raise ValueError("The record key '" + name + "' is not an identier")
            if not isinstance(value, float) and not isinstance(value, str) and not isinstance(value, int):
                raise ValueError("A record must be int, float or string")
    

    def __str__(self):
        return self.table + " " + str(self.values)
    

    def is_empty(self) -> bool:
        return len(self.values) == 0


    def __hash__(self) -> int:
        # see https://stackoverflow.com/a/5884123
        return hash(self.table) + hash(frozenset(self.values.items()))


    def subsetOf(self, other: Record):
        """
        This record is subsetOf other if other includes all name/values of this.
        """
        match = True
        for name, value in self.values.items():
            if name not in other.values or other.values[name] != value:
                match = False
        return match





from __future__ import annotations
from dataclasses import dataclass
from richard.type.PositionType import PositionType


@dataclass(frozen=True)
class RuleConstituent:
    predicate: str
    arguments: list[str]
    position_type: PositionType

    def equals(self, other: RuleConstituent):
        if self.predicate != other.predicate:
            return False
        
        if len(self.arguments) != len(other.arguments):
            return False
        
        if self.position_type != other.position_type:
            return False

        return True
    


from richard.entity.ParseTreeNode import ParseTreeNode
from richard.entity.ProcessResult import ProcessResult
from richard.entity.SentenceRequest import SentenceRequest
from richard.interface.SomeProcessor import SomeProcessor
from richard.interface.SomeSemanticComposer import SomeSemanticComposer


class SemanticExecutor(SomeProcessor):
    """
    Executes the function that forms the meaning of the sentence, and produces its result
    """
    
    composer: SomeSemanticComposer


    def __init__(self, composer: SomeSemanticComposer) -> None:
        super().__init__()
        self.composer = composer    

    
    def process(self, request: SentenceRequest) -> ProcessResult:
        semantic_function = self.composer.get_semantic_function(request)
        results = [semantic_function()]
        return ProcessResult(results, "", [])


    def get_results(self, request: SentenceRequest) -> list:
        return request.get_current_product(self)




import re
from richard.constants import TERMINAL, CATEGORY_PROPER_NOUN, NO_SENTENCE, NOT_UNDERSTOOD, POS_TYPE_REG_EXP, POS_TYPE_RELATION, POS_TYPE_WORD_FORM, UNKNOWN_WORD
from richard.entity.GrammarRule import GrammarRule
from richard.entity.GrammarRules import GrammarRules
from richard.entity.Log import Log
from richard.entity.ProcessResult import ProcessResult
from richard.entity.RuleConstituent import RuleConstituent
from richard.interface.SomeLogger import SomeLogger
from .entity.Chart import Chart
from .entity.ChartState import ChartState
from .tree_extract import extract_tree_roots
from .unknown_word import find_unknown_word


class EarleyParser:
    """ 
    An implementation of Earley's top-down chart parsing algorithm as described in
    "Speech and Language Processing" (first edition) - Daniel Jurafsky & James H. Martin (Prentice Hall, 2000)
    """

    log: SomeLogger

    def __init__(self, log: SomeLogger = Log(False)):
        self.log = log


    def parse(self, grammar_rules: GrammarRules, tokens: list[str]) -> ProcessResult:

        chart = self.buildChart(grammar_rules, tokens, "s", ["P"])

        rootNodes = extract_tree_roots(chart)
        error = ""
        error_args = []

        if len(rootNodes) == 0:

            nextWord = find_unknown_word(chart)

            if nextWord != "":
                error = UNKNOWN_WORD
                error_args = [nextWord]
            elif len(tokens) == 0:
                error = NO_SENTENCE
                error_args = []
            else:
                error = NOT_UNDERSTOOD
                error_args = []

        return ProcessResult(
            products=rootNodes,
            error_code=error,
            error_args=error_args,
        )
    

    def buildChart(self, grammar_rules: GrammarRules, words, rootCategory, rootVariables):
        """
        The body of Earley's algorithm
        """

        chart = Chart(words, rootCategory, rootVariables)
        wordCount = len(words)

        chart.enqueue(chart.build_incomplete_gamma_state(), 0)

        for i in range(wordCount + 1):

            j = 0
            while j < len(chart.states[i]):

                # a state is a is_complete entry in the chart (rule, dot_position, start_word_index, end_word_index)
                state = chart.states[i][j]

                # check if the entry is parsed completely
                if not state.is_complete():

                    # add all entries that have this abstract consequent as their antecedent
                    self.predict(grammar_rules, chart, state)

                    # if the current word in the sentence has this part-of-speech, then
                    # we add a completed entry to the chart (part-of-speech => word)
                    if i < wordCount:
                        self.scan(chart, state)
                else:

                    # proceed all other entries in the chart that have this entry's antecedent as their next consequent
                    self.complete(chart, state)

                j += 1

        return chart
    

    def predict(self, grammar_rules: GrammarRules, chart: Chart, state: ChartState):
        """
        Adds all entries to the chart that have the current consequent of $state as their antecedent.
        """

        consequentIndex = state.dot_position - 1
        nextConsequent = state.rule.consequents[consequentIndex]
        next_consequent_variables = state.rule.consequents[consequentIndex].arguments
        end_word_index = state.end_word_index

        if self.log.is_active():
            self.log.add_debug("predict", state.to_string(chart))

        for rule in grammar_rules.find_rules(nextConsequent.predicate, len(next_consequent_variables)) :

            predicted_state = ChartState(rule, 1, end_word_index, end_word_index)
            chart.enqueue(predicted_state, end_word_index)

            if self.log.is_active():
                self.log.add_debug("> predicted", predicted_state.to_string(chart))


    def scan(self, chart: Chart, state: ChartState):
        """
        If the current consequent in state (which non-abstract, like noun, verb, adjunct) is one
        of the parts of speech associated with the current word in the sentence,
        then a new, completed, entry is added to the chart: (cat => word)
        """

        next_consequent = state.rule.consequents[state.dot_position - 1]
        next_pos_type = state.rule.consequents[state.dot_position - 1].position_type
        next_variables = state.rule.consequents[state.dot_position - 1].arguments
        end_word_index = state.end_word_index
        end_word = chart.words[end_word_index]
        lex_item_found = False
        new_pos_type = POS_TYPE_RELATION

        if self.log.is_active():
            self.log.add_debug("scan", state.to_string(chart))

        if next_pos_type == POS_TYPE_REG_EXP:
            if re.match(next_consequent, end_word):
                lex_item_found = True
                new_pos_type = POS_TYPE_REG_EXP

        # proper noun
        if not lex_item_found and next_consequent == CATEGORY_PROPER_NOUN:
            lex_item_found = True

        # literal word form
        if not lex_item_found:
            if (next_consequent.predicate == end_word.lower()) and (len(next_variables) == 0):
                lex_item_found = True
                new_pos_type = POS_TYPE_WORD_FORM

        if lex_item_found:
            rule = GrammarRule(
                RuleConstituent(next_consequent.predicate, next_variables, new_pos_type),
                [RuleConstituent(end_word, [TERMINAL], POS_TYPE_WORD_FORM)],
                None
            )

            scanned_state = ChartState(rule, 2, end_word_index, end_word_index+1)
            chart.enqueue(scanned_state, end_word_index+1)

            if self.log.is_active():
                self.log.add_debug("> scanned", scanned_state.to_string(chart)+" "+end_word)


    def complete(self, chart: Chart, completed_state: ChartState):
        """
        This function is called whenever a state is complete.
        Its purpose is to advance other states.
        
        For example:
        - this state is NP -> noun, it has been completed
        - now proceed all other states in the chart that are waiting for an NP at the current position
        """

        completed_antecedent = completed_state.rule.antecedent.predicate

        if self.log.is_active():
            self.log.add_debug("complete", completed_state.to_string(chart))

        # index the completed state for fast lookup in the tree extraction phase
        chart.index_completed_state(completed_state)

        for charted_state in chart.states[completed_state.start_word_index] :

            dot_position = charted_state.dot_position
            rule = charted_state.rule

            if (dot_position > len(rule.consequents)) or (rule.consequents[dot_position-1].predicate != completed_antecedent):
                continue

            # check if the types match
            if charted_state.rule.consequents[dot_position-1].position_type != completed_state.rule.antecedent.position_type:
                continue

            # create a new state that is a dot-advancement of an older state
            advanced_state = ChartState(rule, dot_position+1, charted_state.start_word_index, completed_state.end_word_index)

            # enqueue the new state
            chart.enqueue(advanced_state, completed_state.end_word_index)

            if self.log.is_active():
                self.log.add_debug("> advanced", advanced_state.to_string(chart))



from abc import ABC, abstractmethod

from richard.entity.ProcessResult import ProcessResult


"""
Every processor needs to have access to the current alternative interpretation of its predecessor.
To this end it must inject these processors as dependencies. The dependency serves as a key: 
    only with the key can the pipeline be sure that the dependent processor exists.
"""
class SomeProcessor(ABC):
    @abstractmethod
    def process(self, request) -> ProcessResult: # sorry, can't annotate type, because of circular dependency
        pass



from abc import ABC, abstractmethod


class SomeLogger(ABC):
    @abstractmethod
    def is_active(self):
        pass

    @abstractmethod
    def add_debug(self, code: str, text: str):
        pass



from richard.store.Record import Record
from richard.store.RecordSet import RecordSet


class MemoryDb:
    """
    A simple volatile data store. It indexes Records per table.
    """

    store: dict[str, list[Record]]


    def __init__(self) -> None:
        self.store = {}


    def insert(self, record: Record):
        if not record.table in self.store:
            self.store[record.table] = []

        self.store[record.table].append(record)


    def delete(self, record: Record):
        if not record.table in self.store:
            return

        records = []
        for r in self.store[record.table]:
            if not record.subsetOf(r):
                records.add(r)
        self.store[record.table] = records


    def select(self, record: Record) -> RecordSet:
        """
        returns all records from record's table that include record
        """
        result = RecordSet()

        if record.table in self.store:
            for r in self.store[record.table]:
                if record.subsetOf(r):
                    result.add(r)
        
        return result
    



from richard.entity.ProcessResult import ProcessResult
from richard.entity.SentenceRequest import SentenceRequest
from richard.interface.SomeLanguageSelector import SomeLanguageSelector


class LanguageSelector(SomeLanguageSelector):
    """
    Just picks one of the available locales, one by one. If processing the input with one locale fails, the next locale is tried.
    Always used in combination with the `Multilingual` composite processor.
    """

    locales: list[str]
    
    def __init__(self, locales: list[str]) -> None:
        super().__init__()
        self.locales = locales


    def process(self, request: SentenceRequest) -> ProcessResult:
        return ProcessResult(self.locales, "", [])
    

    def get_locale(self, request: SentenceRequest) -> str:
        return request.get_current_product(self)



from richard.constants import GAMMA, POS_TYPE_RELATION
from richard.entity.GrammarRule import GrammarRule
from richard.entity.RuleConstituent import RuleConstituent
from .ChartState import ChartState


class Chart:
    root_category: str
    root_variables: list[str]
    words: list[str]
    states: list[list[ChartState]]
    # all states that were completed, indexed by end word index
    completed_states: dict[int, list[ChartState]]


    def __init__(self, words: list[str], root_category: str, root_variables: list[str]) -> None:
        self.root_category = root_category
        self.words = words
        self.root_variables = root_variables
        self.states = [[] for _ in range(len(words) + 1)]
        self.completed_states = {}

    
    def build_incomplete_gamma_state(chart):
        return ChartState(
            GrammarRule(
                RuleConstituent(GAMMA, ["G"], POS_TYPE_RELATION),
                [RuleConstituent(chart.root_category, chart.root_variables, POS_TYPE_RELATION)],
                None,
            ),
            1, 0, 0)


    def build_complete_gamma_state(chart):
        return ChartState(
            GrammarRule(
                RuleConstituent(GAMMA, ["G"], POS_TYPE_RELATION),
                [RuleConstituent(chart.root_category, chart.root_variables, POS_TYPE_RELATION)],
                None,
            ),
            2, 0, len(chart.words))


    def enqueue(chart, state, position):
        found = chart.contains_state(state, position)
        if not found:
            chart.states[position].append(state)

        return found


    def contains_state(chart, state, position):
        for present_state in chart.states[position] :
            if present_state.equals(state):
                return True

        return False        

    
    def index_completed_state(self, completed_state: ChartState):
        if not completed_state.end_word_index in self.completed_states:
            self.completed_states[completed_state.end_word_index] = []
        self.completed_states[completed_state.end_word_index].append(completed_state)



import unittest

from richard.Pipeline import Pipeline
from richard.block.FindOne import FindOne
from richard.entity.SentenceRequest import SentenceRequest
from richard.processor.parser.BasicParser import BasicParser
from richard.processor.semantic_composer.SemanticComposer import SemanticComposer
from richard.processor.tokenizer.BasicTokenizer import BasicTokenizer

class TestComposer(unittest.TestCase):
   
    def test_missing_sem(self):
        grammar = [
            { "syn": "s -> proper_noun verb", "sem": lambda proper_noun, verb: "not a function" },
            { "syn": "proper_noun -> 'mary'" },
            { "syn": "verb -> 'walks'" },
        ]

        tokenizer = BasicTokenizer()
        parser = BasicParser(grammar, tokenizer)
        composer = SemanticComposer(parser)

        pipeline = Pipeline([
            FindOne(tokenizer),
            FindOne(parser),
            FindOne(composer),
        ])

        exception_occurred = False
        try:
            pipeline.enter(SentenceRequest("Mary walks"))
        except Exception as e:
            self.assertEqual(str(e), "Rule 'proper_noun -> 'mary'' is missing key 'sem'")
            exception_occurred = True
        
        self.assertEqual(exception_occurred, True)


    def test_inner_not_a_function(self):
        grammar = [
            { "syn": "s -> proper_noun verb", "sem": lambda proper_noun, verb: "this is not a function" },
            { "syn": "proper_noun -> 'mary'", "sem": lambda: lambda: None },
            { "syn": "verb -> 'walks'", "sem": lambda: lambda: None },
        ]

        tokenizer = BasicTokenizer()
        parser = BasicParser(grammar, tokenizer)
        composer = SemanticComposer(parser)

        pipeline = Pipeline([
            FindOne(tokenizer),
            FindOne(parser),
            FindOne(composer),
        ])

        exception_occurred = False
        try:
            pipeline.enter(SentenceRequest("Mary walks"))
        except Exception as e:
            self.assertEqual(str(e), "Rule 's -> proper_noun verb' key 'sem' does not return a function")
            exception_occurred = True
        
        self.assertEqual(exception_occurred, True)


    def test_outer_not_a_function(self):
        grammar = [
            { "syn": "s -> proper_noun verb", "sem": "this is not a function" },
            { "syn": "proper_noun -> 'mary'", "sem": lambda: lambda: None },
            { "syn": "verb -> 'walks'", "sem": lambda: lambda: None },
        ]

        tokenizer = BasicTokenizer()
        parser = BasicParser(grammar, tokenizer)
        composer = SemanticComposer(parser)

        pipeline = Pipeline([
            FindOne(tokenizer),
            FindOne(parser),
            FindOne(composer),
        ])

        exception_occurred = False
        try:
            pipeline.enter(SentenceRequest("Mary walks"))
        except Exception as e:
            self.assertEqual(str(e), "Rule 's -> proper_noun verb' key 'sem' is not a function")
            exception_occurred = True
        
        self.assertEqual(exception_occurred, True)



from richard.entity.BlockResult import BlockResult
from richard.entity.SentenceRequest import SentenceRequest
from richard.entity.ControlBlock import ControlBlock


class Succeed(ControlBlock):
    """
    This blocks is added as a terminal, so that the other blocks don't need to check if there's a next one
    """

    def process(self, request: SentenceRequest) -> BlockResult:
        return BlockResult('', [])

    


from richard.store.Record import Record


class RecordSet:
    """
    A record set is a set of records: each record can appear only once 
    """
    
    # each record can occur only once
    _records: set[Record]


    def __init__(self) -> None:
        self._records = set()


    def __len__(self) -> int:
        return len(self._records)


    def add(self, record: Record):
        self._records.add(record)


    # make the set iterable
    def __iter__(self):
        for record in self._records:
            yield record

    
    def field(self, field: str) -> list:
        """
        Returns a list with all distinct values of field
        """
        values = set()
        for record in self._records:
            if field in record.values:
                values.add(record.values[field])
        return list(values)



from dataclasses import dataclass

from richard.block.Succeed import Succeed
from richard.entity.BlockResult import BlockResult
from richard.entity.ControlBlock import ControlBlock
from richard.entity.SentenceRequest import SentenceRequest


@dataclass(frozen=True)
class Pipeline:
    """
    This class contains all blocks/processors involved in the processing of a sentence.
    The number and type of blocks and processors depend on the purpose of the application.
    When a new sentence request comes in, it is sent through the blocks.
    Each processor generates one or more alternative interpretations (ambiguity), and the pipeline processes these one by one.
    Each processor has access to the active alternative interpretations of its predecessors.
    """
    
    blocks: list[ControlBlock]
    block_index: int = -1

    def __post_init__(self):
        # this blocks is added last, so that the other blocks don't need to check if there's a next one
        terminal = Succeed(None)
        for i, block in enumerate(self.blocks):
            next_block = self.blocks[i+1] if i < len(self.blocks)-1 else terminal
            block.next_block = next_block


    def enter(self, request: SentenceRequest) -> BlockResult:
        return self.blocks[0].process(request)




import unittest

from richard.Domain import Domain
from richard.entity.Entity import Entity


class TestPostgresDB(unittest.TestCase):
    """
    To run this test, install psycopg2, set up a PostgreSQL database named "richard",
    create a table "customer" (with columns id and name) and fill it with (1, 'John') (2, 'Jack')
    """
   
    def test_postgres(self):


        # skip for now
        return

        try:
            import psycopg2
            from psycopg2.extras import RealDictCursor

            connection = psycopg2.connect(
                database='richard',
                host='127.0.0.1',
                user='patrick',
                password='test123',
                port=5432,
                cursor_factory=RealDictCursor
            )
        except ImportError:
            raise Exception("To run this test, import psycopg")
        
        def get_all_ids(table: str):
            cursor = connection.cursor()
            cursor.execute("SELECT id FROM " + table)
            return [row['id'] for row in cursor.fetchall()]

        domain = Domain([
            Entity('customer', lambda: get_all_ids('customer'))
        ], [])

        self.assertEqual(domain.get_entity_ids('customer'), [1, 2])

        



import unittest

from richard.Pipeline import Pipeline
from richard.block.FindOne import FindOne
from richard.entity.SentenceRequest import SentenceRequest
from richard.processor.parser.BasicParser import BasicParser
from richard.processor.tokenizer.BasicTokenizer import BasicTokenizer

class TestParser(unittest.TestCase):
   
    def test_parser_process(self):

        grammar = [
            { "syn": "s -> np vp" },
            { "syn": "vp -> verb np" },
            { "syn": "np -> noun" },
            { "syn": "noun -> proper_noun" },
            { "syn": "proper_noun -> 'john'" },
            { "syn": "proper_noun -> 'mary'" },
            { "syn": "verb -> 'loves'" },
        ]

        tokenizer = BasicTokenizer()
        parser = BasicParser(grammar, tokenizer)

        pipeline = Pipeline([
            FindOne(tokenizer),
            FindOne(parser)
        ])

        request = SentenceRequest("John loves Mary")
        pipeline.enter(request)

        tree = parser.get_tree(request)
        self.assertEqual(tree.inline_str(), "s(np(noun(proper_noun(john 'John'))) vp(verb(loves 'loves') np(noun(proper_noun(mary 'Mary')))))")
    

    def test_syntax_error(self):
        grammar = [
            { "syn": "s => proper_noun verb" },
            { "syn": "proper_noun -> 'mary'" },
            { "syn": "verb -> 'walks'" },
        ]

        tokenizer = BasicTokenizer()

        try:
            parser = BasicParser(grammar, tokenizer)
        except Exception as e:
            self.assertEqual(str(e), "Could not parse 'syn' value: s => proper_noun verb")



import unittest

from richard.Pipeline import Pipeline
from richard.block.FindAll import FindAll
from richard.block.FindOne import FindOne
from richard.entity.SentenceRequest import SentenceRequest
from richard.processor.parser.BasicParser import BasicParser
from richard.processor.semantic_composer.SemanticComposer import SemanticComposer
from richard.processor.semantic_executor.SemanticExecutor import SemanticExecutor
from richard.processor.tokenizer.BasicTokenizer import BasicTokenizer

class TestExecutor(unittest.TestCase):

    def test_find_all(self):

        grammar = [
            { "syn": "s -> 'what' 'is' term", "sem": lambda expr: lambda: expr() },
            { "syn": "s -> 'calculate' term", "sem": lambda expr: lambda: expr() },
            { 
            "syn": "term -> term operator term", 
            "sem": lambda term1, operator, term2: 
                        lambda: operator(term1, term2) 
            },
            { "syn": "operator -> 'plus'",  "sem": lambda: lambda a, b: a() + b() },
            { "syn": "operator -> 'minus'", "sem": lambda: lambda a, b: a() - b() },
            { "syn": "operator -> 'times'",  "sem": lambda: lambda a, b: a() * b() },
            { "syn": "operator -> 'divided' 'by'", "sem": lambda: lambda a, b: a() / b() },
            { "syn": "term -> 'one'", "sem": lambda: lambda: 1 },
            { "syn": "term -> 'two'", "sem": lambda: lambda: 2 },
            { "syn": "term -> 'three'", "sem": lambda: lambda: 3 },
            { "syn": "term -> 'four'", "sem": lambda: lambda: 4 },
        ]

        tokenizer = BasicTokenizer()
        parser = BasicParser(grammar, tokenizer)
        composer = SemanticComposer(parser)
        executor = SemanticExecutor(composer)

        pipeline = Pipeline([
            FindOne(tokenizer),
            FindOne(parser),
            FindOne(composer),
            FindOne(executor)
        ])

        request = SentenceRequest("What is three plus four")
        pipeline.enter(request)
        self.assertEqual(executor.get_results(request), 7)
        
        # test find_all
        # test ambiguous sentence with 2 readings (that exposed an error in the parser, now solved)

        pipeline = Pipeline([
            FindOne(tokenizer),
            FindAll(parser),
            FindAll(composer),
            FindAll(executor)
        ])
        request = SentenceRequest("Calculate three plus four times two")
        pipeline.enter(request)

        self.assertEqual(request.get_alternative_products(executor), [14, 11])



from abc import abstractmethod

from richard.entity.SentenceRequest import SentenceRequest
from richard.interface.SomeProcessor import SomeProcessor


class SomeLanguageSelector(SomeProcessor):
    @abstractmethod
    def get_locale(self, request: SentenceRequest) -> str: 
        pass



from richard.entity.ParseTreeNode import ParseTreeNode
from richard.entity.ProcessResult import ProcessResult
from richard.entity.SentenceRequest import SentenceRequest
from richard.interface.SomeParser import SomeParser
from richard.interface.SomeSemanticComposer import SomeSemanticComposer


class SemanticComposer(SomeSemanticComposer):
    """
    Performs semantic composition on the product of the parser
    """
    
    parser: SomeParser


    def __init__(self, parser: SomeParser) -> None:
        super().__init__()
        self.parser = parser    

    
    def process(self, request: SentenceRequest) -> ProcessResult:
        parse_tree = self.parser.get_tree(request)
        semantic_function = self.compose_semantics(parse_tree)
        return ProcessResult([semantic_function], "", [])
    

    def compose_semantics(self, node: ParseTreeNode) -> callable:

        # when using the semantic composer, each rule needs a 'sem'
        if node.rule.sem is None:
            raise Exception("Rule '" + node.rule.basic_form() + "' is missing key 'sem'")

        if not callable(node.rule.sem):
            raise Exception("Rule '" + node.rule.basic_form() + "' key 'sem' is not a function")

        # collect the semantic functions of the child nodes
        child_semantics = []
        for child in node.children:
            if child.form == "":
                semantic_function = self.compose_semantics(child)               
                child_semantics.append(semantic_function)

        # create the semantics of this node by executing its (outer) semantic function, passing the 
        # functions of its children as arguments
        #
        # if you're porting this construct to a language that doesn't support list expansion, 
        # create a switch with a case for each number of arguments (up to 10 or so)
        semantics = node.rule.sem(*child_semantics)

        if not callable(semantics):
            raise Exception("Rule '" + node.rule.basic_form() + "' key 'sem' does not return a function")

        return semantics
    

    def get_semantic_function(self, request: SentenceRequest) -> callable:
        return request.get_current_product(self)



from richard.interface.SomeProcessor import SomeProcessor


class SentenceRequest:
    """
    A request is the data structure that keeps track of the active interpretation of the sentence.
    This interpretation consists of one alternative interpretation per processor so far.
    """

    # raw text that serves as the input to the request
    text: str

    # each processor creates one or more products; the pipeline considers these one by one; the current product is kept here
    current_products: dict[SomeProcessor, any]

    # all alternative products (more than one product suggests ambiguity)
    alternative_products: dict[SomeProcessor, list[any]]
    

    def __init__(self, text: str) -> None:
        self.text = text
        self.current_products = {}
        self.alternative_products = {}


    def set_alternative_products(self, processor: SomeProcessor, alternatives: list[any]):
        self.alternative_products[processor] = alternatives


    def get_alternative_products(self, processor: SomeProcessor) -> list[any]:
        if processor in self.alternative_products:
            return self.alternative_products[processor]
        else:
            return []


    def set_current_product(self, processor: SomeProcessor, alternative: any):
        self.current_products[processor] = alternative


    def get_current_product(self, processor: SomeProcessor):
        if processor in self.current_products:
            return self.current_products[processor]
        else:
            raise Exception("No product available for " + processor.__class__.__name__)



import unittest

from richard.Pipeline import Pipeline
from richard.block.FindOne import FindOne
from richard.entity.SentenceRequest import SentenceRequest
from richard.processor.language_selector.LanguageSelector import LanguageSelector
from richard.processor.language_selector.Multilingual import Multilingual
from richard.processor.parser.BasicParser import BasicParser
from richard.processor.tokenizer.BasicTokenizer import BasicTokenizer

class TestMultilingual(unittest.TestCase):
   
    def test_two_languages(self):

        language_selector = LanguageSelector(["en_US", "nl_NL"])
        tokenizer = BasicTokenizer()

        parsers = {
            "nl_NL": BasicParser([
                { "syn": "s -> np vp" },
                { "syn": "vp -> verb np" },
                { "syn": "np -> noun" },
                { "syn": "noun -> proper_noun" },
                { "syn": "proper_noun -> 'john'" },
                { "syn": "proper_noun -> 'mary'" },
                { "syn": "verb -> 'loves'" },
            ], tokenizer),
            "en_US": BasicParser([
                { "syn": "s -> np vp" },
                { "syn": "vp -> verb np" },
                { "syn": "np -> noun" },
                { "syn": "noun -> proper_noun" },
                { "syn": "proper_noun -> 'jan'" },
                { "syn": "proper_noun -> 'marie'" },
                { "syn": "verb -> 'houdt' 'van'" },
            ], tokenizer)
        }

        parser = Multilingual(parsers, language_selector)

        pipeline = Pipeline([
            FindOne(language_selector),
            FindOne(tokenizer),
            FindOne(parser)
        ])

        request = SentenceRequest("John loves Mary")

        pipeline.enter(request)
        tree = parser.get_product(request)
        self.assertEqual(tree.inline_str(), "s(np(noun(proper_noun(john 'John'))) vp(verb(loves 'loves') np(noun(proper_noun(mary 'Mary')))))")

        request = SentenceRequest("Jan houdt van Marie")

        pipeline.enter(request)
        tree = parser.get_product(request)
        self.assertEqual(tree.inline_str(), "s(np(noun(proper_noun(jan 'Jan'))) vp(verb(houdt 'houdt' van 'van') np(noun(proper_noun(marie 'Marie')))))")
        


from abc import abstractmethod

from richard.entity.SentenceRequest import SentenceRequest
from richard.interface.SomeProcessor import SomeProcessor


"""
Every processor needs to have access to the current alternative interpretation of its predecessor.
To this end it must inject these processors as dependencies. The dependency serves as a key: 
    only with the key can the pipeline be sure that the dependent processor exists.
"""
class SomeTokenizer(SomeProcessor):
    @abstractmethod
    def get_tokens(self, request: SentenceRequest) -> list[str]: 
        pass



import itertools
from richard.entity.ParseTreeNode import ParseTreeNode
from .entity.Chart import Chart
from .entity.ChartState import ChartState


def extract_tree_roots(chart: Chart):

    gamma_state = chart.build_complete_gamma_state()
    if not gamma_state.end_word_index in chart.completed_states:
        return []
    
    gamma_nodes = create_trees_for_state(chart, gamma_state)
    return [gamma_node.children[0] for gamma_node in gamma_nodes]


def create_trees_for_state(chart: Chart, state: ChartState):
    trees = []

    if state.is_terminal():
        trees.append(ParseTreeNode(
            state.rule.antecedent.predicate,
            [],
            state.rule.consequents[0].predicate,
            state.rule
        ))
    else:
        # the state has one or more child states. in fact, there can be multiple child state sequences
        for child_state_sequence in find_child_state_sequences(chart, state):
            # each child state can have multiple parse trees, built from its children; go through all permutations of these child state parse trees
            for permutation_trees in create_trees_for_states(chart, child_state_sequence):
                trees.append(ParseTreeNode(
                    state.rule.antecedent.predicate,
                    permutation_trees,
                    "",
                    state.rule
                ))

    return trees


def find_child_state_sequences(chart: Chart, parent_state: ChartState) -> list[list[ChartState]]:
    """
    each state may have multiple sequences of possible children, find them.
    """
    return find_state_sequences(chart, parent_state, parent_state.end_word_index, len(parent_state.rule.consequents) - 1)


def find_state_sequences(chart: Chart, parent_state: ChartState, end_word_position: int, consequent_index: int):
    """
    try to match the consequent_index'th consequent of parent_state to a new state, ending in the end_word_position
    and prepend the earlier states. this may result in multiple state sequences
    """
    sequences = []
    for state in chart.completed_states[end_word_position]:
        if state.rule.antecedent.equals(parent_state.rule.consequents[consequent_index]):
            if consequent_index == 0 and state.start_word_index == parent_state.start_word_index:
                sequences.append([state])
            else:
                # find one or more preceding sequences of this state
                for previous_sequence in find_state_sequences(chart, parent_state, state.start_word_index, consequent_index-1):
                    sequences.append(previous_sequence + [state])
    return sequences


def create_trees_for_states(chart: Chart, states: list[ChartState]) -> list[list[ParseTreeNode]]:
    # each state has multiple trees
    trees_per_state = [create_trees_for_state(chart, state) for state in states]
    # create all of their permutations
    return create_permutations(trees_per_state)


def create_permutations(things: list[list]) -> list[list]:
    """
    For a list [[a, b], [1, 2], [W, X]], create all permutations, like [a, 1, W] and [b, 1, X]
    See for example https://www.geeksforgeeks.org/python-all-possible-permutations-of-n-lists/
    """
    return list(itertools.product(*things))



from dataclasses import dataclass

from richard.constants import POS_TYPE_RELATION, POS_TYPE_WORD_FORM
from richard.entity.RuleConstituent import RuleConstituent


@dataclass(frozen=True)
class GrammarRule:
    
    antecedent: RuleConstituent
    consequents: list[RuleConstituent]
    sem: callable


    def equals(self, other_rule) -> bool:

        if not self.antecedent.equals(other_rule.antecedent):
            return False

        if len(self.consequents) != len(other_rule.consequents):
            return False
        
        for i, consequent in enumerate(self.consequents):
            if not consequent.equals(other_rule.consequents[i]):
                return False
            
        return True
    

    def basic_form(self):

        s = self.antecedent.predicate #+ "("
        # sep2 = ""
        # for variable in self.antecedent.arguments:
        #     s += sep2 + variable
        #     sep2 = ", "
        # s += ")"

        s += " -> "

        sep = ""
        for i in range(len(self.consequents)):
            if self.consequents[i].position_type == POS_TYPE_RELATION:
                s += sep + self.consequents[i].predicate # + "("
                # sep2 = ""
                # for variable in self.consequents[i].arguments :
                #     s += sep2 + variable
                #     sep2 = ", "
                # s += ")"
            elif self.consequents[i].position_type == POS_TYPE_WORD_FORM:
                s += sep + "'" + self.consequents[i].predicate + "'"
            else:
                s += sep + "/" + self.consequents[i].predicate + "/"
            sep = " "

        return s




TERMINAL = "terminal"
GAMMA = "gamma"

NAME_NOT_FOUND = "Name not found: %s"
SORT_CONFLICT = "It doesn't make sense"
WHICH_ONE = "Which one?"
UNKNOWN_WORD = "Unknown word: %s"
NO_SENTENCE = "No sentence given"
NOT_UNDERSTOOD = "Sorry, I don't understand"
NO_RESULTS = "No results"

POS_TYPE_RELATION = "relation"
POS_TYPE_WORD_FORM = "word-form"
POS_TYPE_REG_EXP = "reg-exp"

CATEGORY_TEXT = "text"
CATEGORY_PROPER_NOUN = "proper_noun"
CATEGORY_PROPER_NOUN_GROUP = "proper_noun_group"



from richard.constants import POS_TYPE_WORD_FORM
from richard.entity.GrammarRule import GrammarRule


class GrammarRules:

    index: dict[str, dict[int, list[GrammarRule]]]


    def __init__(self, grammar_rules: list[GrammarRule]):
        self.index = {}
        for rule in grammar_rules:
            self.add_rule(rule)


    def add_rule(self, rule: GrammarRule):

        antecedent = rule.antecedent.predicate
        argument_count = len(rule.antecedent.arguments)

        if not antecedent in self.index:
            self.index[antecedent] = {}
        if not argument_count in self.index[antecedent]:
            self.index[antecedent][argument_count] = []

        self.index[antecedent][argument_count].append(rule)


    def merge(self, otherRules):
        for m in otherRules.index :
            for r in m :
                for rule in r :
                    self.add_rule(rule)


    def find_rules(self, antecedent: str, argument_count: int) -> list[GrammarRule]:
        if antecedent in self.index and argument_count in self.index[antecedent]:
            return self.index[antecedent][argument_count]
        else:
            return []


    def import_from(self, rules):
        for rules_per_category in rules.index :
            for rulesPerCount in rules_per_category :
                for rule in rulesPerCount :
                    self.add_rule(rule)



from __future__ import annotations
from dataclasses import dataclass
from richard.entity.GrammarRule import GrammarRule


@dataclass(frozen=True)
class ParseTreeNode:

    category: str
    children: list[ParseTreeNode]
    form: str
    rule: GrammarRule

    def is_leaf_node(self) -> bool:
        return len(self.children) == 0
    

    def __str__(self, indent: str = ""):
        body = ""

        if indent == "":
            body = self.category + "\n"

        for i, child in enumerate(self.children):
            if child.form != "":
                body += indent + "+- " + child.category + " '" + child.form + "'\n"
                continue

            body += indent + "+- " + child.category + "\n"
            newIndent = indent
            if i < len(self.children)-1:
                newIndent += "|  "
            else:
                newIndent += "   "
            body += child.__str__(newIndent)

        return body
    
    def inline_str(self):
        body = ""
        sep = ""
        for i, child in enumerate(self.children):
            body += sep
            if child.form != "":
                body += child.category + " '" + child.form + "'"
            else:
                body += child.inline_str()
            sep = " "

        return self.category + "(" + body + ")"
    
    
        


from dataclasses import dataclass
from richard.constants import TERMINAL
from richard.entity.GrammarRule import GrammarRule


@dataclass(frozen=True)
class ChartState:

    rule: GrammarRule
    dot_position:    int
    start_word_index: int
    end_word_index:   int
    
    
    def is_terminal(self):
        if len(self.rule.consequents[0].arguments) == 0:
            return False
        return self.rule.consequents[0].arguments[0] == TERMINAL


    def is_complete(self):
        return self.dot_position >= len(self.rule.consequents) + 1


    def equals(self, other_state):
        return self.rule.equals(other_state.rule) and \
               self.dot_position == other_state.dot_position and \
               self.start_word_index == other_state.start_word_index and \
               self.end_word_index == other_state.end_word_index


    def to_string(self, chart):
        s = self.rule.antecedent.predicate + " ->"
        for i, consequent in enumerate(self.rule.consequents):
            if i + 1 == self.dot_position:
                s += " *"
            s += f" {consequent.predicate}"
        if len(self.rule.consequents) + 1 == self.dot_position:
            s += " *"

        s += " <"
        for i, word in enumerate(chart.words):
            if i >= self.start_word_index and i < self.end_word_index:
                s += f" {word}"
        s += " >"
        return s




from abc import abstractmethod

from richard.entity.SentenceRequest import SentenceRequest
from richard.interface.SomeProcessor import SomeProcessor

class SomeSemanticExecutor(SomeProcessor):
    @abstractmethod
    def get_results(self, request: SentenceRequest) -> list: 
        pass



import unittest

from richard.entity.SentenceRequest import SentenceRequest
from richard.processor.tokenizer.BasicTokenizer import BasicTokenizer

class TestTokenizer(unittest.TestCase):
   
    def test_tokenizer(self):

        tokenizer = BasicTokenizer()

        tests = [
            { "input": "John walks home.", "expected": ["John", "walks", "home", "."]},
            { "input": "Did John's dog walk home?", "expected": ["Did", "John", "'", "s", "dog", "walk", "home", "?"]},
            { "input": "vacuum_cleaner, vacuum-cleaner", "expected": ["vacuum_cleaner", ",", "vacuum", "-", "cleaner"]},
        ]

        for test in tests:
            result = tokenizer.process(SentenceRequest(test["input"]))
            self.assertEqual(result.products[0], test["expected"])

    



from richard.entity.BlockResult import BlockResult
from richard.entity.SentenceRequest import SentenceRequest
from richard.entity.ControlBlock import ControlBlock


class TryFirst(ControlBlock):
    """
    This block just tries the first of the alternative products. If it failes, no other products are tried.
    """
    

    def process(self, request: SentenceRequest) -> BlockResult:
        result = self.processor.process(request)
        if result.error_code != '':
            return BlockResult(result.error_code, result.error_args)

        request.set_alternative_products(self.processor, result.products)

        error_code = ""
        error_args = []

        for product in result.products:

            request.set_current_product(self.processor, product)
          
            next_block_result = self.next_block.process(request)
            error_code = next_block_result.error_code
            error_args = next_block_result.error_args

            # first product tried: quit
            break

        return BlockResult(error_code, error_args)

    


from dataclasses import dataclass

from richard.entity.Relation import Relation
from richard.entity.Entity import Entity


class Domain:
    """
    This class represents the state of the world, and is the source of truth for the pipeline
    """

    indexed_entities: dict[str, Entity]
    indexed_relations: dict[str, Relation]


    def __init__(self, entities: list[Entity], relations: list[Relation]) -> None:
        self.indexed_entities = {}
        for type in entities:
            self.indexed_entities[type.name] = type

        self.indexed_relations = {}
        for relation in relations:
            self.indexed_relations[relation.name] = relation


    def get_entity_ids(self, type_name: str):
        if not type_name in self.indexed_entities:
            raise Exception('No type ' + type_name + " in domain")
        
        type = self.indexed_entities[type_name]
        return type.get_all_ids()


    def relation_exists(self, type_name: str, field_values: list[any]):
        if not type_name in self.indexed_relations:
            raise Exception('No relation ' + type_name + " in domain")
        
        relation = self.indexed_relations[type_name]
        return relation.get(*field_values)



import re

from richard.constants import POS_TYPE_RELATION, POS_TYPE_WORD_FORM
from richard.entity.GrammarRule import GrammarRule
from richard.entity.GrammarRules import GrammarRules
from richard.entity.RuleConstituent import RuleConstituent
from richard.type.SimpleGrammar import SimpleGrammar


class SimpleGrammarRulesParser:
    re_rule: re.Pattern
    re_space: re.Pattern


    def __init__(self) -> None:
        self.re_rule = re.compile("\s*(\w+)\s*->((\s*[\S]+)+)")
        self.re_space = re.compile("\s+")


    def parse(self, simple_grammar: SimpleGrammar):
        rules = []

        for simple_rule in simple_grammar:
            if not 'syn' in simple_rule:
                raise Exception("A rule must contain a 'syn' value")
            syntax = simple_rule['syn']

            result = self.re_rule.match(syntax)
            if not result:
                raise Exception("Could not parse 'syn' value: " + syntax)            
            
            simple_antecedent = result.group(1)
            simple_consequents = re.split(self.re_space, result.group(2).strip())

            antecedent = RuleConstituent(simple_antecedent, ["X"], POS_TYPE_RELATION)
            consequents = []
            for simple_consequent in simple_consequents:
                if simple_consequent[0] == "'":
                    simple_consequent = simple_consequent[1:-1]
                    consequents.append(RuleConstituent(simple_consequent, [], POS_TYPE_WORD_FORM))
                else:
                    consequents.append(RuleConstituent(simple_consequent, ["X"], POS_TYPE_RELATION))

            sem = None
            if 'sem' in simple_rule:
                sem = simple_rule['sem']
            rules.append(GrammarRule(antecedent, consequents, sem))

        return GrammarRules(rules)
    


from abc import abstractmethod

from richard.entity.SentenceRequest import SentenceRequest
from richard.interface.SomeProcessor import SomeProcessor


class SomeSemanticComposer(SomeProcessor):
    @abstractmethod
    def get_semantic_function(self, request: SentenceRequest) -> callable: 
        pass



from setuptools import setup, find_packages

setup(
    name='richard',
    version='0.2.0',    
    description='A Natural Language Understanding and Execution library',
    url='https://github.com/garfix/richard',
    author='Patrick van Bergen',
    author_email='patrick.vanbergen@gmail.com',
    license='MIT',
    packages=find_packages(),
    install_requires=[],

    classifiers=[
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',  
        'Programming Language :: Python :: 3.10',
    ],
)




def find(np: tuple[callable, callable], vp: callable) -> list:
    """
    Result consists of all elements in np.nbar that satisfy vp
    If the number of results agrees with qp, results are returned. If not, an empty list
    """
    qp, nbar = np
    elements = nbar()
    range_count = len(elements)

    result = []
    for element in elements:
        for e2 in vp(element):
            result.append(element)

    result = list(set(result))
    result_count = len(result)
    
    if qp(result_count, range_count):
        return result
    else:
        return []
    


import unittest

from richard.Domain import Domain
from richard.Pipeline import Pipeline
from richard.block.FindOne import FindOne
from richard.entity.Entity import Entity
from richard.entity.Relation import Relation
from richard.store.Record import Record
from richard.entity.SentenceRequest import SentenceRequest
from richard.processor.parser.BasicParser import BasicParser
from richard.processor.semantic_composer.SemanticComposer import SemanticComposer
from richard.processor.semantic_executor.SemanticExecutor import SemanticExecutor
from richard.processor.tokenizer.BasicTokenizer import BasicTokenizer
from richard.semantics.commands import find
from richard.store.MemoryDb import MemoryDb


class TestQuantifier(unittest.TestCase):
   
    def test_quantifier(self):

        db = MemoryDb()
        db.insert(Record('has_child', {'parent': 'mary', 'child': 'lucy'}))
        db.insert(Record('has_child', {'parent': 'mary', 'child': 'bonny'}))
        db.insert(Record('has_child', {'parent': 'barbara', 'child': 'john'}))
        db.insert(Record('has_child', {'parent': 'barbara', 'child': 'chuck'}))
        db.insert(Record('has_child', {'parent': 'william', 'child': 'oswald'}))
        db.insert(Record('has_child', {'parent': 'william', 'child': 'bertrand'}))

        domain = Domain([
            Entity("parent", lambda: db.select(Record('has_child')).field('parent')),
            Entity("child", lambda: db.select(Record('has_child')).field('child')),
        ], [
            Relation("has_child", ['parent', 'child'], lambda parent, child: db.select(Record('has_child', {'parent': parent, 'child': child}))),
        ])

        grammar = [
            { 
                "syn": "s -> np vp_no_sub", 
                "sem": lambda np, vp_no_sub: lambda: find(np(), vp_no_sub) 
            },
            { 
                "syn": "vp_no_sub -> aux qp child", 
                "sem": lambda aux, qp, child:
                        lambda subject: find(
                            (qp, child),
                            lambda object: domain.relation_exists('has_child', [subject, object]))
            },
            { 
                "syn": "np -> qp nbar", 
                "sem": lambda qp, nbar:  lambda: (qp, nbar) 
            },
            { 
                "syn": "qp -> det", 
                "sem": lambda det: lambda result, range: det(result, range) 
            },
            { "syn": "nbar -> noun", "sem": lambda noun: lambda: noun() },
            { 
                "syn": "det -> 'every'", 
                "sem": lambda: lambda result, range: result == range 
            },
            { 
                "syn": "det -> number", 
                "sem": lambda number: lambda result, range: result == number() 
            },
            { "syn": "number -> 'two'", "sem": lambda: lambda: 2 },
            { "syn": "number -> 'three'", "sem": lambda: lambda: 3 },
            { "syn": "noun -> 'parent'", "sem": lambda: lambda: domain.get_entity_ids('parent') },
            { "syn": "child -> 'children'", "sem": lambda: lambda: domain.get_entity_ids('child') },
            { "syn": "aux -> 'has'", "sem": lambda: lambda: None },
        ]

        tokenizer = BasicTokenizer()
        parser = BasicParser(grammar, tokenizer)
        composer = SemanticComposer(parser)
        executor = SemanticExecutor(composer)

        pipeline = Pipeline([
            FindOne(tokenizer),
            FindOne(parser),
            FindOne(composer),
            FindOne(executor)
        ])

        request = SentenceRequest("Every parent has two children")
        pipeline.enter(request)
        results = executor.get_results(request)
        self.assertEqual(len(results), 3)

        request = SentenceRequest("Every parent has three children")
        pipeline.enter(request)
        results = executor.get_results(request)
        self.assertEqual(len(results), 0)



import re
from richard.entity.ProcessResult import ProcessResult
from richard.entity.SentenceRequest import SentenceRequest
from richard.interface.SomeTokenizer import SomeTokenizer


class BasicTokenizer(SomeTokenizer):

    BASIC_TOKEN_RE = "([a-zA-Z_0-9]+|[^\\s])"
    
    token_expression: re.Pattern

    def __init__(self) -> None:
        super().__init__()
        self.token_expression = re.compile(self.BASIC_TOKEN_RE)

    def process(self, request: SentenceRequest) -> ProcessResult:
        tokens = self.token_expression.findall(request.text)
        return ProcessResult([tokens], "", [])


    def get_tokens(self, request: SentenceRequest) -> list[str]:
        return request.get_current_product(self)



from richard.constants import NO_RESULTS
from richard.entity.BlockResult import BlockResult
from richard.entity.SentenceRequest import SentenceRequest
from richard.entity.ControlBlock import ControlBlock


class FindAll(ControlBlock):
    """
    This block goes through all alternative products even after one was found that gives no error.
    If no product could be found, however, the first error is returned.
    """

    def process(self, request: SentenceRequest) -> BlockResult:
        result = self.processor.process(request)
        if result.error_code != '':
            return BlockResult(result.error_code, result.error_args)

        error_code = ""
        error_args = []

        # the essence of this block: collect all products
        request.set_alternative_products(self.processor, 
            request.get_alternative_products(self.processor) + result.products)

        success = False
        for product in result.products:

            request.set_current_product(self.processor, product)

            next_block_result = self.next_block.process(request)
            success = success or next_block_result.error_code == ""
            if next_block_result.error_code != "" and error_code == "":
                error_code = next_block_result.error_code
                error_args = next_block_result.error_args
           
        if success:
            return BlockResult("", [])    
        else:
            return BlockResult(error_code, error_args)
    



from .entity.Chart import Chart


# Returns the word that could not be parsed (or ""), and the index of the last completed word
def find_unknown_word(chart: Chart):
    """
    Returns the word that could not be parsed (or ""), and the index of the last completed word
    """

    nextWord = ""
    last_understood_index = -1

    # for i = len(chart.states) - 1; i >= 0; i -= 1:
    for i in range(len(chart.states) - 1, -1, -1):
        states = chart.states[i]
        for state in states :
            if state.is_complete():
                if state.end_word_index > last_understood_index:
                    last_understood_index = state.end_word_index - 1

    if last_understood_index+1 < len(chart.words):
        nextWord = chart.words[last_understood_index+1]
    else:
        nextWord = " ".join(chart.words)

    return nextWord



from richard.entity.GrammarRules import GrammarRules
from richard.entity.ParseTreeNode import ParseTreeNode
from richard.entity.ProcessResult import ProcessResult
from richard.entity.SentenceRequest import SentenceRequest
from richard.interface.SomeParser import SomeParser
from richard.interface.SomeTokenizer import SomeTokenizer
from richard.processor.parser.helper.SimpleGrammarRulesParser import SimpleGrammarRulesParser
from richard.type.SimpleGrammar import SimpleGrammar
from .earley.EarleyParser import EarleyParser


class BasicParser(SomeParser):

    tokenizer: SomeTokenizer
    grammar: GrammarRules
    parser: EarleyParser


    def __init__(self, grammar: SimpleGrammar, tokenizer: SomeTokenizer) -> None:

        self.tokenizer = tokenizer

        sgrp = SimpleGrammarRulesParser()
        self.grammar = sgrp.parse(grammar)
        self.parser = EarleyParser()
            

    def process(self, request: SentenceRequest) -> ProcessResult:
        tokens = self.tokenizer.get_tokens(request)
        return self.parser.parse(self.grammar, tokens)
    

    def get_tree(self, request: SentenceRequest) -> ParseTreeNode:
        return request.get_current_product(self)
    



from richard.entity.ProcessResult import ProcessResult
from richard.entity.SentenceRequest import SentenceRequest
from richard.interface.SomeLanguageSelector import SomeLanguageSelector
from richard.interface.SomeProcessor import SomeProcessor


class Multilingual(SomeProcessor):
    """
    This compositing processor takes other processors as input and uses the processor of the active language when asks to `process` input
    It can be used to combine any combinations of processors, like tokenization processors or parsing processors
    """
    
    processors: dict[str, SomeProcessor]
    language_selector: SomeLanguageSelector

    def __init__(self, processors: dict[str, SomeProcessor], language_selector: SomeLanguageSelector) -> None:
        super().__init__()
        self.processors = processors
        self.language_selector = language_selector

        
    def process(self, request: SentenceRequest) -> ProcessResult:

        locale = self.language_selector.get_locale(request)

        if locale not in self.processors:
            raise Exception("No processor available for locale " + locale)

        return self.processors[locale].process(request)
    

    def get_product(self, request: SentenceRequest) -> any:
        return request.get_current_product(self)
    
